""" plugin_dbui.py

    Controllers expose by the plugin:
    
        index
        database
        configuration

    Author: R. Le Gac

"""

import os


API = """
Ext.namespace('App');
App.csvUrl = '/%s/plugin_dbui/csv';
App.name = '%s';
App.REMOTE_API = {
    'url': '/%s/plugin_dbui/call',
    'type': 'remoting',
    'actions': %s 
};
App.storeCfgs = %s; 
App.viewportCfg = %s; """


MATHJAX = """
MathJax.Hub.Config({
    extensions: ["tex2jax.js"],
    jax: ["input/TeX", "output/HTML-CSS"],
    tex2jax: {
        inlineMath: [ ['$','$'], ["\\(","\\)"] ],
        displayMath: [ ['$$','$$'], ["\\[","\\]"] ],
        processEscapes: true
    },
    "HTML-CSS": { availableFonts: ["TeX"] }
});
"""

def call():
    """Action to handle the Ext.Direct protocol.
    
    """
    return directSvc()


def csv():
    """Action returning the content of a table as a CSV file.
    Foreign keys are replaced by the parent column.
    
    """
    import gluon.contenttype
    dbui = local_import('plugin_dbui')

    tablename = request.vars['tableName']

    response.headers['Content-Type'] = gluon.contenttype.contenttype('.csv')
    response.headers['Content-Disposition'] = 'attachment; filename=%s.csv;' % tablename
    
    table = db[tablename]
    query = table.id > 0
    
    # replace foreign key by parent fields
    # and build the query to resolve foreign keys
    if dbui.is_table_with_foreign_fields(table):
    
        fields = []
        query = ((query) & (dbui.get_where_query(table)))
        
        for field in table:
        
            if dbui.is_foreign_field(field):
                k_table, k_field, k_key = dbui.get_foreign_field(field)
                fields.append(db[k_table][k_field])
                
            else:
                fields.append(field)
    else:
        fields = [table.ALL]

    # interrogate the database    
    rows = db(query).select(*fields)
    return str(rows)
    

def dbui_conf():
    """Action returning the javascript script configuring
    the plugin dbui: application name, store definition,...
    
    """
    from gluon.contrib import simplejson as json
    dbui = local_import('plugin_dbui')
    di = {}
    
    # build the dictionary required by Ext.direct
    # containing the definition of the remote procedure
    for (k, f) in directSvc.procedures.iteritems():
        action, method = k.split('.')

        if action == dbui.DBUI:
            nargs = f.func_code.co_argcount
        else:
            nargs = f.__func__.func_code.co_argcount - 1
            
        if action not in di:
            di[action] = []
            
        di[action].append({'name': method, 'len': nargs})
  
    # the stores configuration (static, for each table,...)
    storeCfgs = plugins.dbui.static_stores
    for table in db:
        cfg = dbui.to_jsonstore(table)
        storeCfgs[cfg['storeId']] = cfg 
    
    # the viewport configuration
    viewportCfg = dbui.to_viewport()
    
    # fill the javascript template
    app = request.application
    script = API % (app, 
                    app,
                    app, 
                    json.dumps(di),
                    json.dumps(storeCfgs),
                    json.dumps(viewportCfg))
    
    # return the response as a javascript
    response.headers['Content-Type'] = 'text/javascript'
    return script


def debug():
    """Action to run the plugin in debug mode
    Load the javascript source code for all libraries.
    Active the debug mode on the server side.

    EXAMPLES
    
        http://localhost:8000/myapp/plugin_dbui/debug
        http://localhost:8000/myapp/plugin_dbui/debug?script=foo

    URL OPTIONS

        script 
            run the javascript foo in the framework defined by the plugin.
            The scripts are stored in the directory defined by the plugin
            configuration app_script_dir. 

    """
    # check the presence of the ExtJS and MathJax plugins
    dbui = local_import('plugin_dbui')
    plugin = plugins.dbui

    # css files
    response.files.extend(dbui.get_file_paths(plugin.extjs_css, ext='.css'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_css, ext='.css'))
    response.files.extend(dbui.get_file_paths(plugin.app_css, ext='.css'))

    # debug version of the javascript libraries
    response.files.extend(dbui.get_file_paths(plugin.mathjax_libmin, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.extjs_js, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_js, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.app_js, ext='.js'))

    # language files
    response.files.extend(dbui.get_file_paths(plugin.extjs_lg, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_lg, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.app_lg, ext='.js'))
    
    # URL for dbui configuration script
    response.files.append(plugin.dbui_conf)
    
    # main script which can be defined in many different ways
    pscript = dbui.get_script_path(plugin)
    response.files.append(pscript)

    # switch ON the debug mode on the server side
    session.debug = True

    # select the view    
    response.view = 'plugin_dbui/index.html'
    
    # return the plugin configuration parameter as well as the response 
    return dict(plugin=plugin)
    

def index():
    """Default Action to run the plugin
    Load compressed version of all libraries.
    
    EXAMPLES
        
        http://localhost:8000/myapp/plugin_dbui
        http://localhost:8000/myapp/plugin_dbui?script=foo
        
    URL OPTIONS

        script
            run the javascript foo in the framework defined by the plugin.
            The scripts are stored in the directory defined by the plugin
            configuration app_script_dir.

    """
    dbui = local_import('plugin_dbui')
    plugin = plugins.dbui

    # css files
    response.files.extend(dbui.get_file_paths(plugin.extjs_css, ext='.css'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_css, ext='.css'))
    response.files.extend(dbui.get_file_paths(plugin.app_css, ext='.css'))

    # compressed version of the javascript libraries
    response.files.extend(dbui.get_file_paths(plugin.mathjax_libmin, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.extjs_libmin, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_libmin, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.app_libmin, ext='.js'))
    
    # language files
    response.files.extend(dbui.get_file_paths(plugin.extjs_lg, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.dbui_lg, ext='.js'))
    response.files.extend(dbui.get_file_paths(plugin.app_lg, ext='.js'))
    
    # URL for dbui configuration script
    response.files.append(plugin.dbui_conf)
    
    # main script which can be defined in many different ways
    pscript = dbui.get_script_path(plugin)
    response.files.append(pscript)

    # switch off the debug mode on the server side
    session.debug = False

    # return the plugin configuration parameter as well as the response 
    return dict(plugin=plugin)


def mathjax_conf():
    """Action returning the js configuration script for the MathJax library.
    
    """
    return MATHJAX


def status():
    return dict(request=request, response=response, session=session)