""" Converters transforming Web2py DAL objects into ExtJS widgets.

Author: R. Le Gac

"""
import gluon.dal
import re

from extjs import *
from gluon import current
from helper import (encode_field,
                    get_field_validators,
                    get_foreign_field,
                    get_set_field, 
                    is_foreign_field,
                    is_set_field)
from mapper import map_default

# Associated gluon.dal.field type with an ExtJS widget
# The dictionary contains configuration parameters of the Ext JS widget
FTYPE_TO_XTYPE = {'boolean':   (CheckBox, {}),\
                  'date':      (DateField, {'format': 'Y-m-d'}),\
                  'datetime':  (DateField, {'format': 'Y-m-d H:i:s'}),\
                  'double':    (NumberField, {'decimalPrecision':2,
                                              'decimalSeparator': '.'}),\
                  'id':        (TextField, {}),\
                  'integer':   (NumberField, {}),\
                  'password':  (TextField, {'inputType': 'password'}),\
                  'reference': (ComboBox, {}),\
                  'string':    (TextField, {}),\
                  'text':      (TextArea, {}),\
                  'time':      (TimeField, {}),\
                  'upload':    (TextField, {'inputType': 'file'})}


# constant defining a ExtJS store
ROOT = 'records'
STOREID = '%sStore'
SUCCESS= 'success'
TOTAL = 'count'


def _to_field(field, **kwargs):
    """Private converter. It is recommended to used to_field.
    
    Convert a gluon.dal.Field into an ExtJS.form.Field and its
    inherited class. It does not handle composite field.
    The conversion takes into account the FieldsModifier instructions.
    
    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    """
    T = current.T
    
    fieldname = field.name
    tablename = field.tablename

    f_name = encode_field(tablename, fieldname)
                
    # foreign key
    if is_foreign_field(field):
            
        k_tablename, k_fieldname, k_key = get_foreign_field(field)
        
        cfg = ComboBox(name=f_name,
                       hiddenName=f_name,
                       displayField=encode_field(k_tablename, k_fieldname),
                       valueField=encode_field(k_tablename, k_key),
                       store=STOREID % k_tablename)

    # set field
    elif is_set_field(field):
        cfg = SetBox(name=f_name,
                     model={"setData": get_set_field(field)})
        
    # other types of fields
    else:
        cls, di = FTYPE_TO_XTYPE[field.type] 
        cfg = cls(name=f_name, **di)

    # field label translate in the local languqge
    cfg["fieldLabel"] = str(T(field.label))

    # default value and not null
    # NOTE: the entryForm doesn't work when both default and
    # notnull condition are defined.
    if field.default:
        cfg["value"] = str(field.default)
    elif field.notnull:
        cfg["allowBlank"] = False

    # read only field
    if field.readable and not field.writable:
        cfg["readOnly"] = True

    # tool tip
    if field.comment:
        cfg["tipText"] = field.comment
     
    # validator constraints
    cfg.update(get_field_validators(field))
    
    # hide primary key
    # hide field which are not readable and not writable
    # hide field request by the form modifier
    hidden = field.name == "id"
    hidden = hidden or ((not field.readable) and (not field.writable))
    
    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        if fieldname in modifier_forms[tablename].hidden_fields:
            hidden = True
            
    if hidden:
        cfg["hidden"] = True
        cfg["hideLabel"] = True
        cfg["readOnly"] = True

    # configuration options set by the field_modifers
    modifier_fields = PluginManager('dbui').dbui.modifier_fields
    if tablename in modifier_fields:
        if fieldname in modifier_fields[tablename].extjs_fields:
            cfg.update(modifier_fields[tablename].extjs_fields[fieldname])

    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_field(field, composite=True, **kwargs):
    """Convert a gluon.dal.Field into an Ext.form.Field and its inherited class
    as well as into  a composite field, i.e Ext.form.CompositeField.
     
    Composite fields are set using the FieldsModifer.
    Return None if the field is consumed by a CompositeField.
    composite field can be ignore when the keyword argument composite is False.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.

    """
    table = field.table
    tablename = field.tablename
    modifier_fields = PluginManager('dbui').dbui.modifier_fields

    # do we have composite field for this table ?
    composite_fields = None
    if tablename in modifier_fields:
        modifier_field = modifier_fields[tablename]
        composite_fields = modifier_field.composite_fields

    # main field of the composite field
    # it will consume the embedded field too
    if composite and composite_fields and field.name in composite_fields.main:
            
        cfg = CompositeField()

        for fieldname in composite_fields[field.name].fields:
            cfg.append_items(_to_field(table[fieldname]))
        
        # use configuration options from the modifier
        cfg.update(composite_fields[field.name].extjs)
    
    # an embedded field already used in a composite field
    elif composite and composite_fields and field.name in composite_fields.others:
        cfg = None
    
    # a basic field
    else:
        cfg = _to_field(field)

    # configuration options from keyword arguments
    if cfg:
        cfg.update(kwargs)
    
    return cfg

    
def to_fields(table):
    """Convert a gluon.dal.Table into a list of Ext.form.Field,  
    Ext.form.CompositeField or and Ext.form.FieldSet. 

    Individual fields can be organized in Fieldset and in CompositeField.
    They are defined in the application model using modifiers namely
    the FormModifier and the FieldsModifier respectively.

    The conversion takes into account the modifier instructions.
    
    """
    li = []
    tablename = table._tablename
    modifier_forms =  PluginManager('dbui').dbui.modifier_forms

    # do we have FieldSets
    field_sets, modifier_form = None, None
    if tablename in modifier_forms:
        modifier_form = modifier_forms[tablename]
        if modifier_form and modifier_form.field_sets:
            field_sets = modifier_form.field_sets
            
    # Table with Ext.form.FieldSet and/or Ext.form.CompositeFields 
    # and/or Ext.form.Field and/or Spacer

    # NOTE: the field id is add in order to run this form
    # in all context, create, update, destroy.
    # It is add to the latest fieldset.
    id_is_used = False 

    if field_sets:
        for fieldset in field_sets:
            
            cfg = FieldSet()            
            for fieldname in fieldset.fields:
                        
                if isinstance(fieldname, Base):
                    cfg.append_items(fieldname)
                    continue

                if fieldname == 'id':
                    id_is_used = True

                field = table[fieldname]
                item = to_field(field)
                if item:
                    cfg.append_items(item)
        
            # fieldset configuration options ser by the modifier
            cfg.update(fieldset.extjs)
            li.append(cfg)
            
        # append the field Id the last field set
        if not id_is_used:
            li[-1].append_items(to_field(table['id']))

    # Table with Ext.form.CompositeFields and/or Ext.form.Fields
    else:
        for field in table:
            cfg = to_field(field)
            if cfg:
                li.append(cfg)
    
    # map the list of fields/fieldSets on Ext.form.formPanel
    mapper = map_default
    if modifier_form and modifier_form.mapper:
        mapper = modifier_form.mapper

    li = mapper(li)

    return li


def to_formPanel(table, **kwargs):
    """Convert a gluon.dal.Table into a ExtJS App.form.FormPanel.
    The conversion takes into account the form modifier instructions.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    """
    tablename = table._tablename

    fields = to_fields(table)
    cfg = FormPanel(items=fields)

    # add a store for non dummy database        
    if table._db._uri:
        cfg['store'] = STOREID % tablename
    
    # configuration options from form modifier
    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        cfg.update(modifier_forms[tablename].extjs)
        
    # configuration options form the keyword arguments
    cfg.update(kwargs)
    
    return cfg 


def to_gridColumn(field, **kwargs):
    """Convert a gluon.dal.Field into an Ext.grid.Column.
    The conversion takes into account the grid modifier instructions.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
            
    """
    T = current.T
    
    fieldname = field.name
    tablename = field.tablename
    
    cfg = GridColumn(header=str(T(field.label)),
                     sortable=True)
    
    # name of the field in the data store
    # replace foreign key by the pointing column
    if is_foreign_field(field):
        k_tablename, k_fieldname, k_key = get_foreign_field(field)
        index = encode_field(k_tablename, k_fieldname)

    else:
        index  = encode_field(tablename, fieldname)

    cfg["dataIndex"] = index
        
    # hide the primary key
    if fieldname == "id":
        cfg["hidden"] = True

    # hide field which are not readable and not writable
    if (not field.readable) and (not field.writable):
        cfg["hidden"] = True

    # Hide fields request by the grid modifiers
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].hidden_columns:
            cfg["hidden"] = True
    
    # configuration options from the grid modifier
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].configure_columns:
            cfg.update(modifier_grids[tablename].configure_columns[fieldname])
    
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_gridColumnModel(table):
    """Convert a gluon.dal.Table into an Ext.grid.ColumnModel.
    The conversion takes into account the grid modifier instructions.
    
    Currently the gridColumModel is implemented as a list of gridColumn.
    
    """
    delete_columns = []
    template_columns = []
    tablename = table._tablename

    # get modifier requirements
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        delete_columns = modifier_grids[tablename].delete_columns
        template_columns = modifier_grids[tablename].template_columns

    # standard column
    cfg = GridColumnModel()  
    for field in table:
        fieldname = field.name
        if fieldname not in delete_columns:
            cfg.append(to_gridColumn(field))
    
    # template column
    for tpl in template_columns:
        col = GridTemplateColumn(sortable=True)
        col.update(tpl.extjs)
        cfg.insert(tpl.position, col)

    # row numbering in the first column
    if modifier_grids[tablename].row_numbering:
        cfg.insert(0, GridRowNumberer())
        
    return cfg


def to_gridFilter(table, **kwargs):
    """Convert a gluon.dal.Table into an ExtJS App.grid.GridFilter.
    The grid filter is set using the grid modifiers.
    Return an empty dictionary if not defined
    
    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    FILTER RULE:
        A tuple of 3 string (field1, operator, comment).
        Field1 is the name of a field belonging to the table.
        Valid operators are defined in Dbsvc._encode_query.
    
        A more elaborate filter rule allows to filter on any foreign field.
        the syntax is (table2.field1, operator, comments).
    
        Rules are store in modifier section grid_filters.
        
    """    
    T = current.T
    tablename = table._tablename
     
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename not in modifier_grids:
        return {}
        
    grid_filters = modifier_grids[tablename].grid_filters
    if not grid_filters:
        return {}

    if not grid_filters.filters:
        return {}
    
    di = GridFilter(title=T('Filter %s' % tablename))
    
    for i in range(len(grid_filters.filters)):
        
        fieldname, operator, comment = grid_filters.filters[i]
        
        # filter rule dealing with any foreign field
        m = re.match(r"(\w+)\.(\w+)", fieldname)
        if m:
            ktablename = m.group(1)
            kfieldname = m.group(2)
            field = table._db[ktablename][kfieldname]
            name = '[%s.%s]' % (ktablename, kfieldname)
            
        # basic filter rule
        else:
            field = table[fieldname]
            name = '[%s.%s]' % (tablename, fieldname)
        
        # get the standard configuration for the field
        # ignore composite field defined for form
        cfg = to_field(field, composite=False)
        
        # remove default value 
        # make no sense in this context
        if 'value' in cfg:
            del cfg['value']

        # replace all textarea by textfield
        if cfg['xtype'] == 'xtextarea':
            cfg['xtype'] = 'xtextfield'

        # empty text for combobox
        if cfg['xtype'] == 'xcombobox':
            cfg['emptyText']= T('select...')
        
        # prepare the name for the where close [table.field]
        cfg['name'] = name    
            
        # store information to customize the filter
        cfg['filterComment'] = comment  
        cfg['filterOperator'] = operator
        cfg['filterType'] = field.type
                    
        # copy the comment to be used as a tool tip
        # when the plugin pFormToolTip is used
        cfg['tipText'] = comment
        
        # field configuration option
        cfg.update(grid_filters.extjs_filters[i])
        
        di.append_items(cfg)
    
    # configuration options from the grid modifier
    di.update(grid_filters.extjs)
    
    # configuration option from keyword arguments
    di.update(kwargs)
    
    return di
    

def to_gridPanel(table, **kwargs):
    """Convert a gluon.dal.Table into an ExtJS App.grid.GridPanel.
    The conversion takes into account the grid modifier instructions.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    """
    tablename = table._tablename

    cfg = GridPanel(columns=to_gridColumnModel(table),
                    frame=False,
                    viewConfig={'forceFit': True})
    
    # add a store for non dummy database        
    if table._db._uri:
        cfg['store'] = STOREID % tablename
    
    # configuration option from the grid modifier
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        cfg.update(modifier_grids[tablename].extjs)

    # configuration option from the keyword arguments
    cfg.update(kwargs)
    
    # grid with filter
    filter = to_gridFilter(table)
    if filter:
        cfg = GridWithFilter(panelCfg=dict(cfg),
                             selectorCfg=filter)
    
    return cfg


def to_jsonstore(table, **kwargs):
    """Convert a gluon.dal.Table into an App.data.DirectStore.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    """
    db = table._db
    tablename = table._tablename

    # NOTE: the configuration option baseParams defined the
    # transactions parameters seen by the server
    # This dictionary contains 3 keys:
    #    tablename: name of the table in the database
    #    dbFields: list of fields which should be red
    #    where: define inner join, mainly to resolve foreign key
    base_params = {'tableName': tablename, 'dbFields': []}
    
    cfg = DirectStore(autoLoad=False,
                      autoSave=True,
                      baseParams=base_params,
                      fields=[],
                      idProperty=encode_field(tablename,'id'),
                      root=ROOT,
                      successProperty=SUCCESS,
                      storeId=STOREID % tablename,
                      totalProperty=TOTAL)

    for field in table:        
        fieldname = field.name

        # the field of the table
        # encode field name and define default value
        field_cfg = {'name': encode_field(tablename, fieldname)}
        if field.default:
            field_cfg['defaultValue'] = str(field.default)
        
        base_params['dbFields'].append([tablename, fieldname])
        cfg['fields'].append(field_cfg)
        
        # NOTE: the store contains all fields of the table.
        # It also contains the pointing values for foreign keys
        # Therefore the store has the the id and the value for 
        # each foreign key. This technique is use to resolve foreign key
        # in form (combobox) and in grid.
        if is_foreign_field(field):
            
            if 'where' not in base_params: base_params['where'] = []
            k_tablename, k_fieldname, k_key = get_foreign_field(field)
            
            # encode the field name
            k_field_cfg = {'name': encode_field(k_tablename, k_fieldname)}
            base_params['dbFields'].append([k_tablename, k_fieldname])
            cfg['fields'].append(k_field_cfg)
            
            # retrieve the default value for the foreign key
            if field.default:
                row = db(db[k_tablename].id == field.default).select().first()
                if row:
                    k_field_cfg['defaultValue'] = row[k_fieldname]
            
            # define the where close
            nt = (tablename, fieldname, k_tablename, k_key)
            base_params['where'].append("[%s.%s] == [%s.%s]" % nt)
    
    # configuration options from the store modifier
    # namely the order by directive
    modifier_stores = PluginManager('dbui').dbui.modifier_stores
    if tablename in modifier_stores:
        
        base_params['orderby'] = []
        for el in modifier_stores[tablename].orderby:

            if isinstance(el, gluon.dal.Field):
                field, dir = el, 'ASC'
                
            elif isinstance(el, gluon.dal.Expression):
                field, dir = el.first, 'DESC'

            base_params['orderby'].append([field.tablename, field.name, dir])
            
            
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_tree(node):
    """Convert the Node structure into a list of dictionary
    configuring Ext.tree.TreeNode.

    The application navigation tree is defined using ViewportModifier.
    
    The argument node is required by ExtJS 3.4 but it is not
    used here.
    
    """
    cfg = []
    
    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        nav_nodes = modifier_viewports.nodes
        for node in nav_nodes:
            cfg.append(node.get_node())

    return cfg


def to_viewport(**kwargs):
    """Return the configuration dictionary for the App.Viewport.

    ExtJS configuration parameters are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    """
    cfg = {}
    
    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        cfg = modifier_viewports.extjs
        
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg