""" Author: R. Le Gac

"""

import re

from basesvc import BaseSvc
from converter import ROOT, SUCCESS, TOTAL
from gluon.contrib import simplejson as json
from helper import (encode_field, 
                    decode_field,
                    get_foreign_field,
                    is_foreign_field,
                    rows_serializer)


FIELD_NOT_IN_DB = "the field '%s.%s' doesn't exist in the database."
KEYWORD_MISSING = "The keyword '%s' is missing."
RECORD_DELETED = "Record %s is deleted."
RECORD_INSERTED = "Record is inserted with id %i."
RECORD_NOT_IN_DB = "The record '%s' doesn't exist in the table %s."
RECORD_UPDATED = "Record %s is updated."
TABLE_NOT_IN_DB = "The table '%s' doesn't exist in the database."
TO_MANY_RECORDS = "To many records !!!"


class DbSvcException(BaseException): pass


class DbSvc(BaseSvc):
    """Interface the database and client request.
    It is the glue between wep2py and ExtJS library.
    It relies on the Ext.Direct protocol exposing dbSvc methods on
    the client-side.

    Main methods are:

        create
            to create a new record in a table.

        read
            read the content of a a table
            resolving foreign key.

        update
            update the record of a table.

        destroy
            delete the record of a table.
        
    Each method return a dictionary:
    {success: True, records:{blabla}, msg: 'blalbla'}
    
    The keyword success is always present. It is mandatory for 
    the FormPanel widget of the ExtJS library, but not strictly 
    required since failure can be propagated through HTTP errors.
    
    In the records dictionary database field are encode as TableField.

    If the action failed an Exception is raised.
    The exception can be catch to generate an HTTP error.

    """
    def __init__(self, environment):
        
        BaseSvc.__init__(self, environment)
        db = self.environment['db']


    def _check_request(self, arg, keywords=('tableName', 'records')):
        """Helper function to check that the request arg is well formed.
        The dictionary arg contains the following keys:

            tablename
                the name of the table in the database
                
            dbFields
                The list of table fields.

            records (optional)
                A dictionary containing the new / update values for field
                as well as the identifier of the record to be create, updated
                or delete.

        Raise the DbSvcException if a keyword is missing or if the table
        does not exist in the database.
        
        """
        # check that the request is well formed
        for kword in keywords:
            if kword not in arg:
                raise DbSvcException(KEYWORD_MISSING % kword)
        
        table = arg['tableName']
        self._is_table_in_db(table)


    def _encode_query(self, li):
        """Helper method to encode the list of query send by the client
        as a web2py query.
        
                "[table1.field1] == [table2.field2]" or "[table1, field1] == [table2, field2]" 
                "[table1.field1] > n"
                "[table1.field1] like 'm%'"
                "[table1.field1] contains 'm%'"
                "[table1.field1] startswith 'm%'"
                "[table1.field1].lower() like 'm%'"
                "[table1.field1] belongs (1, 2, 3)"
                "([table1.field1] == [table2.field2]) | ([table1.field2] == [table2.field1])"
                
                all element of the list are ANDED
                operator are ==, !=, <, >, <=, >=
                boolean operator are &, |

        """ 
        query = ""
        for el in li:
            
            # transform bracket as db.table1.field1
            s = re.sub(r'\[(\w+).(\w+)\]', r'db.\1.\2', el)
            
            # transform contains 'm%' as .contains("m%')
            for op in ['belongs', 'contains', 'like', 'startswith']:
                if op in s:
                    s = re.sub(r" +(%s) +(.+)" %op, r'.\1(\2)', s)
            
            # build the query
            query = ("(%s) & (%s)" % (query, s) if query else s)

        return eval(query, {}, {"db": self.environment['db']})    

    
    def _get_fields(self, arg):
        """Helper method to get fields and their values in order to
        create or update a record. The dictionary arg contains the following 
        keys:

            tablename
                the name of the table in the database
                
            dbFields
                The list of table fields.
                It also include pointing field resolving foreign keys.
                [(table1, field1), (table1, field2), (table2, field3), ...]

            records (ROOT)
                List of dictionary containing the new / update values for field
                as well as the identifier of the record to be updated (id)

        Return a dictionary {fieldName: value, ...}
        or {"success": False, "errors": {...}, ..} in case of validation errors
        
        """
        fields = {}
        table = arg['tableName']

        # protection
        if len(arg[ROOT]) > 1:
            raise DbSvcException(TO_MANY_RECORDS)
        
        # Remove fields in the record which do not belong to the table.
        # The JsonStore send the foreign key and the the pointing field
        # when manipulating table with foreign keys. the latter belongs
        # to another table.
        record = arg[ROOT][0]
        for table_field in record:
            tablename, fieldname = decode_field(table_field)
            if tablename == table:
                fields[fieldname.encode('utf8')] = record[table_field]

        # Validate field contents
        di = self._is_fields_values_valid(table, fields)
        if di:
            table_id = encode_field(table, 'id')
            record[table_id] = ""
            return {SUCCESS: False, "errors": di, ROOT: record}

        return fields


    def _get_record(self, table, id):
        """Helper function to get the record id in a form which can be decoded
        by the JsonReader running on the client side. Foreign key are resolved.
        Return a dictionary where key are encoded as TableField.
        
        """
        db = self.environment['db']
        record = {}

        for field in db[table].fields:
            key = encode_field(table, field)
            record[key] = db[table][id][field] 

            if is_foreign_field(db[table][field]):
                k_table, k_field, k_query = get_foreign_field(db[table][field])
                
                k_key = encode_field(k_table, k_field)
                k_id = record[key]
                record[k_key] = db[k_table][k_id][k_field]
                
        return record


    def _is_field_in_table(self, table, field):
        """Helper method to check that the field belongs to the table.
        Return a boolean. 
        
        """
        return field in self.environment['db'][table].fields
    
    
    def _is_fields_values_valid(self, table, fields):
        """Helper method to check each field value against its validators.
        Return an empty dictionary when everything is OK or a dictionary
        with fields failing check: {field1: error1, field2: error2,...}
        
        """
        di = {}
        db = self.environment['db']
        
        for (field, val) in fields.items():
            errmsg = db[table][field].validate(val)[1]
            if errmsg:
                di[encode_field(table, field)] = errmsg

        return di


    def _is_table_in_db(self, table):
        """Helper method to check that a table exist in the database.
        Raise the  DbSvcException if not.
        
        """
        if table not in self.environment['db'].tables:
            raise DbSvcException(TABLE_NOT_IN_DB % table)


    def create(self, arg):
        """Create a new record defined in arg.
        The dictionary arg contains the following keys:

            tableName
                the name of the table in the database
                
            dbFields
                The list of table fields.
                It also include pointing field resolving foreign keys.
                [(table1, field1), (table1, field2), (table2, field3), ...]

            records (ROOT)
                A dictionary containing the update values for field
                as well as the identifier of the record to be updated (id)
                
        Return a dictionary with status, message and the update row:
        {success: True, msg: 'blalbla', records:{TableId:xx,....}}

        """
        self.dbg("Start DbSvc.create")

        self._check_request(arg)
        fields = self._get_fields(arg)

        if SUCCESS in fields and fields[SUCCESS] == False:
            return fields

        # insert new record        
        table = arg['tableName']
        id = self.environment['db'][table].insert (**fields)
        
        # return the complete record
        # mandatory, since this is the one which is display in the grid.
        record = self._get_record(table, id)

        self.dbg("End DbSvc.create.", RECORD_INSERTED % id)
        return {SUCCESS: True, 'msg': RECORD_INSERTED % id, ROOT: record}

    
    def destroy(self, arg):
        """Destroy the record defined in arg.
        The dictionary arg contains the following keys:

            tableName
                the name of the table in the database
                
            dbFields
                The list of table fields.
                It also include pointing field resolving foreign keys.
                [(table1, field1), (table1, field2), (table2, field3), ...]

            records (ROOT)
                A list with id numbers
                
        Return a dictionary with status, message and id of the deleted row:
        {success: True, msg: 'blalbla', records:{TableId:xx}}

        """
        self.dbg("Start DbSvc.destroy")

        self._check_request(arg)
        
        db = self.environment['db']
        table = arg['tableName']
        table_id = encode_field(table, 'id')
        
        for id in arg[ROOT]:
            if not db[table][id]:
                return {"success": False, 
                        "msg": RECORD_NOT_IN_DB % (id, table),  
                        ROOT: {table_id: id}}
            
            del db[table][id]
        
        self.dbg("End DbSvc.destroy")
        return {"success": True, "msg": RECORD_DELETED % id,  ROOT: {table_id: id}}
    
    
    def read(self, arg):
        """Read the content of a table as specified in the arg dictionary.
        The latter contains the following keys:
        
            tableName
                name of the database table
                
            dbFields
                A list of field to be read.
                It also include pointing field resolving foreign keys.
                [(table1, field1), (table1, field2), (table2, field3), ...]
                            
            where (optional)
                define the inner join when handling foreign key.
                The syntax follows the web2py convention:
                ['db.table1.field1  == db.table2.id'].
                An AND is performed between the different 
                elements of the list.

            orderby (optional)
                order directive
                A list of fields [(table1, field1, dir), (table2, field3, dir), ...]
                where dir is either ASC or DESC
                
        The dictionary can also contains parameters useful for paging.
        For more detail see  Ext.PagingToolbar.
            start
            limit
            sort
            dir

        The method return a list of records:
        {success: True, records: [{TableField: value, ...}, {...}, ...], ...}
        
        """
        self.dbg("Start DbSvc.read")
        
        self._check_request(arg, ('tableName', 'dbFields'))
        db = self.environment['db']

        # build the list of SQLField object
        fields = []
        for table, field in arg['dbFields']:
            if not self._is_field_in_table(table, field):
                raise DbSvcException(FIELD_NOT_IN_DB % (table, field))
            fields.append(db[table][field])

        # build the inner join conditions
        query = ''
        if 'where' in arg:
            query = self._encode_query(arg['where'])

        # Count the number of record in the table --require for the paging
        if query:
            nrecords = db(query).count()
            
        else:
            table = arg['tableName']
            nrecords = db(db[table].id).count()
        
        # handle paging options (see Ext.PagingToolbar)
        kwargs = {}
        if 'start' in arg and 'limit' in arg:
            start = int(arg['start'])
            limit = int(arg['limit'])
            kwargs['limitby'] = (start, start+limit)

        # handle the orderby directive ['[table1, field2, dir]', ....]
        if 'orderby' in arg:
            kwargs['orderby'] = [] 
            for tablename, fieldname, dir in arg['orderby']:
                el = db[tablename][fieldname]
                if dir == 'DESC':
                    el = ~el
                kwargs['orderby'].append(el)
                     
        # interrogate the database and serialize the records as a list:
        # [{TableField: value, }, {...}, ...]
        rows = db(query).select(*fields, **kwargs)
        li = rows_serializer(rows)
        
        self.dbg('End DbSvc.read', li)
        return {SUCCESS: True, TOTAL: nrecords, ROOT: li}


    def update(self, arg):
        """Update a record according to the arg dictionary.
        The latter contains the following keys:
        
            tableName
                the name of the table in the database
                
            dbFields
                The list of table fields.
                It also include pointing field resolving foreign keys.
                [(table1, field1), (table1, field2), (table2, field3), ...]

            records (ROOT)
                A dictionary containing the update values for field
                as well as the identifier of the record to be updated (id)

        Return a dictionary with status, message and the update row:
        {success: True, msg: 'blalbla', records:{TableId:xx,...}}

        """
        self.dbg("Start DbSvc.update.")

        self._check_request(arg)
        fields = self._get_fields(arg)

        if SUCCESS in fields and fields[SUCCESS] == False:
            return fields

        # the client send modified fields only
        # update the database accordingly        
        id = fields['id']
        table = arg['tableName']
        self.environment['db'][table][id] = fields

        # return the complete record
        # mandatory, since this is the one which is display in the grid.
        record = self._get_record(table, id) 

        self.dbg("End DbSvc.update.")
        return {SUCCESS: True, "msg": RECORD_UPDATED % id, ROOT: record}
