""" ExtJS components configurators 

Author: R. Le Gac

"""

from gluon import current
from gluon.tools import PluginManager

MSG_XTYPE = 'Invalid keyword "xtype"'


class ExtJSException(BaseException): pass


class Base(dict):
    """Base class for ExtJS configurator.
    A protection is set to avoid changing the xtype of the element.

    Two helper functions append_items and append_plugins
    to deal with items and plugins.
    
    """
    xtype = None
    
    def __init__(self, **kwargs):

        if 'xtype' in kwargs:
            raise ExtJSException(MSG_XTYPE)

        self['xtype'] = self.xtype
        self.update(kwargs)


    def _append(self, key, t):
        li = list(t)
        if key in self:
            self[key].extend(li)
        else:
            self[key] = li


    def append_items(self, *args):
        self._append('items', args)

        
    def append_plugins(self, *args):
        self._append('plugins', args)


class ArrayStore(Base):
    """Configurator for the Ext.data.ArrayStore."""
    xtype='arraystore'


class CheckBox(Base):
    """Configurator for Ext.form.Checkbox."""
    xtype = 'checkbox'


class ComboBox(Base):
    """Configurator for App.form.CombBox."""
    xtype = 'xcombobox'


class CompositeField(Base):
    """Configurator for Ext.form.CompositeField."""
    xtype = 'compositefield'


class DateField(Base):
    """Configurator for Ext.form.DateField."""
    xtype = 'datefield'


class DirectStore(Base):
    """Configurator for App.data.DirectStore."""
    xtype = 'xdirectstore'


class Field(Base):
    """Configurator for Ext.form.Field."""
    xtype = 'field'


class FieldSet(Base):
    """Configurator for Ext.form.FieldSet.

    By default, the layout for children Field is set to anchor 99%.
    It can be overwritten using the keywords defaults.
       
    """
    xtype = 'fieldset'
    
    def __init__(self, **kwargs):
        self['defaults'] = {'anchor': '99%'}
        Base.__init__(self, **kwargs)


class FormPanel(Base):
    """Configurator for App.form.FormPanel."""
    xtype = 'xform'


class GridColumn(dict):
    """Configurator for Ext.grid.Column."""


class GridColumnModel(list):
    """Configurator for Ext.grid.ColumnModel implemented 
    as a list of GridColumn.
    
    """


class GridFilter(Base):
    """Configurator for App.grid.GridFilter."""
    xtype = 'xgridfilter'


class GridPanel(Base):
    """Configurator for App.grid.Grid."""
    xtype = 'xgrid'


class GridRowNumberer(Base):
    """Configurator for Ext.grid.RowNumberer."""
    xtype = 'rownumberer'


class GridTemplateColumn(Base):
    """Configurator for Ext.grid.TemplateColumn."""
    xtype = 'templatecolumn'


class GridWithFilter(Base):
    """Configurator for App.grid.GridWithFilter."""
    xtype = 'xgridwithfilter'


class JsonStore(Base):
    """Configurator for the Ext.data.JsonStore."""
    xtype='jsonstore'


class NumberField(Base):
    """Configurator for Ext.form.NumberField."""
    xtype = 'numberfield'


class Panel(Base):
    """Configurator for Ext.Panel."""
    xtype = 'panel'


class PanelWithUrlSelector(Base):
    """Configurator for App.PanelWithUrlSelector."""
    xtype = 'xpanelwithurlselector'


class SetBox(Base):
    """Configurator for App.form.SetBox."""
    xtype = 'xsetbox'


class Spacer(Base):
    """Configurator for Ext.Spacer."""
    xtype = 'spacer'


class Store(Base):
    """Configurator for the Ext.data.Store."""
    xtype='store'


class TabPanel(Base):
    """Configurator for Ext.TabPanel."""
    xtype = 'tabpanel'


class TextArea(Base):
    """Configurator for Ext.form.TextArea."""
    xtype = 'xtextarea'

    
class TextField(Base):
    """Configurator for Ext.form.TextField."""
    xtype = 'xtextfield'

    
class TimeField(Base):
    """Configurator for Ext.form.TimeField."""
    xtype = 'timefield'


class Viewport(Base):
    """Configurator for App.Viewport."""
    xtype = 'xviewport'


class XmlStore(Base):
    """Configurator for the Ext.data.XmlStore."""
    xtype='xmlstore'
    