""" Author: R. Le Gac

"""
import copy

from gluon.storage import Storage
from modifier import Modifier


MODIFIER_GRIDS = 'modifier_grids'

MSG_HEADER_MISSING = 'configuration option header is missing in extjs'
MSG_INVALID_KEYWORD = 'invalid keyword %s'
MSG_POSITION_MISSING = 'keyword argument position is missing'
MSG_TPL_MISSING = 'configuration option tpl is missing in extjs'


def configure_grids(db, **extjs):
    """Helper functions to configure all grid widgets with the same 
    configuration options. Useful to set plugins for example.
    
    The keyword arguments contain the configuration options for
    the Ext.grid.GridPanel.
    
    This tool can be use to set grid plugins for all grids in one go:
    > configure_grids(db, plugins=['pGridRowEditorContextMenu'])
    
    """
    for tablename in db.tables:
        gm = GridModifier(tablename)
        di = copy.deepcopy(extjs)
        gm.configure(**di)


class GridModifierException(BaseException): pass

    
class GridModifier(Modifier):
    """Helper tool to customize the grid widget associated to a table.
    
    """
    
    def __init__(self, tablename):
        """Initialize the modifier persistent data.
        
        """
        Modifier.__init__(self, MODIFIER_GRIDS, tablename)
        
        self.data.configure_columns = {}
        self.data.delete_columns = []
        self.data.grid_filters = Storage(filters=[],
                                         extjs_filters=[], 
                                         extjs={})
        self.data.hidden_columns = []
        self.data.row_numbering = False
        self.data.template_columns = []

                   
    def add_column(self, ktable, kfield, position=0, extjs={}):
        """Add a column displaying the database field kfield of the 
        foreign table ktable. 
        
        The keyword argument position is a number defining where the
        column will be insert. The first column is at 0.
        
        The keyword argument ExtJS is a dictionary containing
        configuration options of the underlying Ext.grid.Column.
        
        """


    def configure_column(self, fieldname, **kwargs):
        """Modified the model of a given column.
        This method is useful to fine tune column width for example.

        The keyword arguments defines the configuration option
        of the underlying Ext.grid.Column

        """
        self.data.configure_columns[fieldname] = kwargs

        
    def delete_columns(self, *fields):
        """Delete the columns in the Ext.grid.GridPanel widget.
        Each column is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.delete_columns:
                self.data.delete_columns.append(el)

        
    def hide_columns(self, *fields):
        """Hide the columns list in arguments
        Each column is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.hidden_columns:
                self.data.hidden_columns.append(el)
        
        
    def merge_columns(self, *fields, **kwargs):
        """Merge columns in a single template column Ext.grid.TemplateColumn.
        Each column is identified by its database field name.
        
        The new column should have an header and a template defining how the 
        merge field are displayed. These two configuration options of the 
        Ext.grid.Templatecolumn have to be defined via the keyword argument 
        extjs. The corresponding key of this dictionary are header and tpl.
        
        The syntax of the template is rather tricky to manipulate. 
        see the documentation of Ext.XTemplate constructor.        
        There is some limitation when defining the template in python.
        For example the use of the template member function is impossible.
        In the template, database fields are identified using the CapWord
        syntax: TablenameFieldName. The first letter of the tablename and field
        name are in upper case while all the other letters are in lower case.
        
        When the keyword argument autohide is set to true the columns in field
        are hidden.
        
        The keyword argument position is a number defining where the
        column will be insert. The first column is at 0.
        
        The keyword arguments extjs defines the configuration option
        of the underlying Ext.grid.TemplateColumn
        
        """
        keywords = ['autohide', 'position']
        extjs = dict(kwargs)
        
        for key in keywords:
            if key in extjs:
                del extjs[key]
        
        if 'position' not in kwargs:
            raise GridModifierException(MSG_POSITION_MISSING)
        
        if 'header' not in kwargs:
            raise GridModifierException(MSG_HEADER_MISSING)

        if 'tpl' not in kwargs:
            raise GridModifierException(MSG_TPL_MISSING)
            
        di = Storage(extjs=extjs, position=kwargs['position'])
        self.data.template_columns.append(di)
        
        if 'autohide' in kwargs and kwargs['autohide']:
            self.hide_columns(*fields)


    def append_filter(self, filter, **kwargs):
        """Append a filter associated to the grid.

        The widget associated to each filter is a field, comboBox, dataField,..
        They are grouped in a Ext.form.Fieldset appearing in a panel on the 
        right side of the grid.
        
        A filter is define by a tuple containing three columns:
            - a string with the database field name
            - a string with the operator. Valid operator are defined in
              the method dbsvc._encode_query
            - a string with a comment for the tool tip
            
        A more elaborate filter rule allows to filter on any foreign field.
        the syntax is (table2.field1, operator, comments).
        
        The keyword argument contains the additional configuration options
        for the underlying Ext.form.Field. For more information see the ExtJS
        documentation.
        
        A set of filters can be defined in one go using the method set_filters.
        This method allows to tune each field while the other allows to tune
        the FieldSet.
        
        """
        self.data.grid_filters.filters.append(filter)
        self.data.grid_filters.extjs_filters.append(kwargs)


    def set_filters(self, *filters, **kwargs):
        """Define the filters associated to the grid.
        
        It is an Ext.form.Fieldset appearing in a panel on the right side 
        of the grid. It contains widget associated to each filter: field, 
        comboBox, dataField,..
        
        A filter is define by a tuple containing three columns:
            - a string with the database field name
            - a string with the operator. Valid operator are defined in
              the method dbsvc._encode_query
            - a string with a comment for the tool tip
        
        A more elaborate filter rule allows to filter on any foreign field.
        the syntax is (table2.field1, operator, comments).
            
        The keyword argument contains the configuration options for the 
        underlying Ext.form.FieldSet. For more information see the ExtJS
        documentation.

        Filter can be defined and tuned individually 
        using the method append_filter.
        
        """
        for el in filters:
            self.append_filter(el)
            
        self.data.grid_filters.extjs=kwargs


    def set_rownumbering(self):
        """Activate the row numbering.
        Row number is displayed in the first column of the grid.
        
        """
        self.data.row_numbering = True
