/**
 * appbase.js
 * 
 * Constants and elementary functions to built an application
 * 
 */
Ext.namespace('App');

/**
 * @cfg {String} App.name
 * Name of the web application.
 * This constants is defined by the server.
 */

/**
 * @cfg {Array} App.storeCfgs
 * Configuration option for all json stores
 */

/**
 * @param {String} App.version version of the library
 */
App.version = '0.4.8.2';

/**
 * Helper function mimicking the encode_field function running on the server
 * (table, field) → TableField
 * @param {String} table
 * @param {String} field
 */
App.encodeField = function (table, field) {
    var t, f;
    t = table[0].toUpperCase() + table.slice(1, table.length);
    f = field[0].toUpperCase() + field.slice(1, field.length);
    return t + f;
};

/**
 * Helper function returning an App.data.Store identifies by its id.
 * If the store does not exit it is created and register in the store manager.
 * 
 * NOTE: The storeId is defined by the server. 
 * The syntax of the storeId is "nameStore". 
 * If the store does not exist, its configuration is extract from 
 * the application constant App.storeCfgs.Then it is created and registered.
 * 
 * In the current frame work, this function allows to associated a unique store
 * to a database table and that widget share the same store.
 * 
 * @param {String} storeId
 */
App.getStore = function (storeId) {
    
    var cfg,
        store,
        table;
    
    store = Ext.StoreMgr.lookup(storeId);
    
    if (!store && !Ext.isString(storeId)) {
        throw new Error('Fail to instanciate the store: "' + storeId + '"');
    }    

    if (!store && Ext.isString(storeId)) {
    
        cfg = App.storeCfgs[storeId];
        
        switch (cfg.xtype) {
            case 'arraystore':
                store = new Ext.data.ArrayStore(cfg);
                break;
                
            case 'directstore':
                store = new Ext.data.DirectStore(cfg);
                break;
                
            case 'jsonstore':
                store = new Ext.data.JsonStore(cfg);
                break;
                
            case 'store':
                store = new Ext.data.Store(cfg);
                break;
                
            case 'xdirectstore':
                store = new App.data.DirectStore(cfg);
                break;
                
            case 'xmlstore':
                store = new Ext.data.XmlStore(cfg);
                break;
                
            default:
                throw new Error('Fail to instanciate the object: "' + cfg.xtype + '"');
                break;
        }
    }
    return store;
};

/**
 * Helper function returning the name of the database table 
 * associated to the store. works with store configured using the
 * Dbui.getJsonStore method.
 * @param {App.data.DirectStore} store
 */
App.getTableName = function (store) {
    return store.baseParams.tableName;    
};

/**
 * Helper function to determine if a plugin is present 
 * in the plugins list of a component
 * @param {Ext.Component} 
 * @param {String} ptype the ptype of the plugin
 */
App.isPlugin = function (component, ptype) {

    var i, 
        plugin;

    if (!component.hasOwnProperty('plugins')) {
        return false;
    }

    for (i = 0; i < component.plugins.length; i += 1) {

        plugin = component.plugins[i];

        if (plugin === ptype) {
            return true;
        }
        
        if ((typeof (plugin) === 'object') && (plugin.ptype === ptype)) {
            return true;
        }   
    }
    return false;
};
