/**
 * The GridFilter class
 * 
 * Associate to grid, it contains all the logic to filter the content of a grid
 * according to the value define in the form.
 * 
 * @version
 *
 */

Ext.namespace('App.grid');

App.grid.GridFilter = Ext.extend(Ext.form.FieldSet, {

    /**
     * Private
     * Dictionary with additional where clause
     */	
    filterConditions: {},
    
    /**
     * Private
     * Initial where clause
     */
    initialFilterConditions: [],
    
    /**
     * Private
     * Reference to the paging toolBar
     */
    pagingToolbar: null,

    /**
     * Private 
     * a reference to the grid store
     */
    store: null,
    
    /**
     * Predefined setting
     * 
     * NOTE: 
     * the enableKeyEvents is mandatory for Ext.form.TextField
     */
    defaults: {
        anchor: '99%',
        enableKeyEvents: true
    },
    
    /**
     * private method requests by the component model of Ext JS
     */
    initComponent: function () {

        var field, 
            fields, 
            i;
        
        // initialize the underlying class.
        App.grid.GridFilter.superclass.initComponent.call(this);
                        
        // associate handlers to fields
        fields = this.findByType('field');
        for (i = 0; i < fields.length; i += 1) {
            field = fields[i];
            
            // catch a change of value for comboBox
            if (field.xtype === 'xcombobox') {
                field.on('select', this.onChange, this);
            
            // catch a change of date
            } else if (field.xtype === 'datefield') {
                field.on('select', this.onChange, this);
                
            // catch the key pressed enter for other fields
            // wait that user finish to type to run the handler
            } else {
                field.on('keyup', this.onChange, this, {buffer: 500});
            }
        }
    },
    
    /**
     * Bind the grid filter to the grid
     * @param {Ext.grid.GridPanel} grid
     */
    bind: function (grid) {

        var baseParams, 
            i;
        
        this.store = grid.getStore();
        
        // paging toolbar ?
        if (App.isPlugin(grid, 'pGridPaging')) {
            this.pagingToolbar = grid.getBottomToolbar();
        }
        
        // initial filter conditions
        baseParams = this.store.baseParams;
        if ('where' in baseParams) {
            for (i = 0; i < baseParams.where.length; i += 1) {
                this.initialFilterConditions.push(baseParams.where[i]);
            }
        }
    },

    /**
     * Handler to catch a change in field value
     */
    onChange: function (field) {
        this.setupCondition(field);
    },

    /**
     * Handler to reset the filter
     */
    onReset: function () {
        
        var fields, 
            i;
        
        // reset the field value
        fields = this.findByType('field');
        for (i = 0; i < fields.length; i += 1) {
            fields[i].reset();
        }
        
        // reset filter conditions
        this.filterConditions = {};
        
        if (this.initialFilterConditions.length > 0) {
            this.store.baseParams.where = this.initialFilterConditions;
                
        } else {
            delete this.store.baseParams.where;
        }
        
        // reset the page size to the default value
        this.pagingToolbar.pageSize = 10;
        
        // update the store
        this.updateStore();
    },
    
    /**
     * Setup condition
     * @param {Ext.form.Field}
     */
    setupCondition: function (field) {

        var conditions = [], 
            i, 
            k, 
            value, 
            where;
        
        // get the field value
        value = field.getValue();
        
        // handle date using the proper format
        if (field.xtype === 'datefield') {
            if (value instanceof Date) {
                value = value.format(field.format);
            } else {
                return;                
            }
        }
         
        //  update the condition dictionary
        where = field.name + " " + field.filterOperator + " '" + value + "'";
        
        if (value === '') {
            delete this.filterConditions[field.name];
        } else {
            this.filterConditions[field.name] = where;
        }
                    
        // build the complete where clause
        for (i = 0; i < this.initialFilterConditions.length; i += 1) {
            conditions.push(this.initialFilterConditions[i]);
        }
        
        for (k in this.filterConditions) {
            if (this.filterConditions.hasOwnProperty(k)) {
                conditions.push(this.filterConditions[k]);
            }
        }
        
        // push new condition in the where clause of the store
        this.store.baseParams.where = conditions;
        
        // update the store
        this.updateStore();
    },
    
    /**
     * private
     * Helper function to load the store applying filter conditions
     */
    updateStore: function () {
        
        var bbar = this.pagingToolbar,
            nRows,
            slider;
        
        if (bbar) {
            nRows = bbar.pageSize;
            slider = bbar.findByType('slider')[0];
            
            this.store.load({params: {start: 0, limit: nRows}, callback: function() {
                bbar.moveFirst();
                slider.setMaxValue(this.store.getTotalCount());
                slider.setValue(nRows);
            }, scope: this});


        } else {
            this.store.load();
        }		
    }
});

Ext.reg('xgridfilter', App.grid.GridFilter);
